@@ -3,11 +3,6 @@
 OpenStack Nova Cloud Module
 ===========================
 
-PLEASE NOTE: This module is currently in early development, and considered to
-be experimental and unstable. It is not recommended for production use. Unless
-you are actively developing code in this module, you should use the OpenStack
-module instead.
-
 OpenStack is an open source project that is in use by a number a cloud
 providers, each of which have their own ways of using it.
 
@@ -50,9 +45,28 @@ examples could be set up in the cloud configuration at
 
       ssh_key_name: mykey
 
-      provider: nova
+      driver: nova
       userdata_file: /tmp/userdata.txt
 
+To use keystoneauth1 instead of keystoneclient, include the `use_keystoneauth`
+option in the provider config.
+
+.. note:: this is required to use keystone v3 as for authentication.
+
+.. code-block:: yaml
+
+    my-openstack-config:
+      use_keystoneauth: True
+      identity_url: 'https://controller:5000/v3'
+      auth_version: 3
+      compute_name: nova
+      compute_region: RegionOne
+      service_type: compute
+      tenant: admin
+      user: admin
+      password: passwordgoeshere
+      driver: nova
+
 For local installations that only use private IP address ranges, the
 following option may be useful. Using the old syntax:
 
@@ -72,7 +86,7 @@ accept them
       user: myusername
       password: mypassword
       tenant: <userid>
-      provider: nova
+      driver: nova
 
     my-api:
       identity_url: 'https://identity.api.rackspacecloud.com/v2.0/'
@@ -81,52 +95,126 @@ accept them
       api_key: <api_key>
       os_auth_plugin: rackspace
       tenant: <userid>
-      provider: nova
+      driver: nova
       networks:
         - net-id: 47a38ff2-fe21-4800-8604-42bd1848e743
         - net-id: 00000000-0000-0000-0000-000000000000
         - net-id: 11111111-1111-1111-1111-111111111111
 
+This is an example profile.
+
+.. code-block:: yaml
+
+    debian8-2-iad-cloudqe4:
+      provider: cloudqe4-iad
+      size: performance1-2
+      image: Debian 8 (Jessie) (PVHVM)
+      script_args: -UP -p python-zmq git 2015.8
+
+and one using cinder volumes already attached
+
+.. code-block:: yaml
+
+    # create the block storage device
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      block_device:
+        - source: image
+          id: <image_id>
+          dest: volume
+          size: 100
+          shutdown: <preserve/remove>
+          bootindex: 0
+
+    # with the volume already created
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      boot_volume: <volume id>
+
+    # create the volume from a snapshot
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      snapshot: <cinder snapshot id>
+
+    # create the create an extra ephemeral disk
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      ephemeral:
+        - size: 100
+          format: <swap/ext4>
+
+    # create the create an extra ephemeral disk
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      swap: <size>
+
+Block Device can also be used for having more than one block storage device attached
+
+.. code-block:: yaml
+
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      block_device:
+        - source: image
+          id: <image_id>
+          dest: volume
+          size: 100
+          shutdown: <preserve/remove>
+          bootindex: 0
+        - source: blank
+          dest: volume
+          device: xvdc
+          size: 100
+          shutdown: <preserve/remove>
+
+Floating IPs can be auto assigned and ssh_interface can be set to fixed_ips, floating_ips, public_ips or private_ips
+
+.. code-block:: yaml
+
+    centos7-2-iad-rackspace:
+      provider: rackspace-iad
+      size: general1-2
+      ssh_interface: floating_ips
+      floating_ip:
+        auto_assign: True
+        pool: public
+
+
 Note: You must include the default net-ids when setting networks or the server
 will be created without the rest of the interfaces
 
 Note: For rackconnect v3, rackconnectv3 needs to be specified with the
-rackconnect v3 cloud network as it's variable
+rackconnect v3 cloud network as its variable.
 '''
-from __future__ import absolute_import
 # pylint: disable=E0102
 
-# The import section is mostly libcloud boilerplate
-
 # Import python libs
+from __future__ import absolute_import
 import os
 import logging
 import socket
 import pprint
 import yaml
 
-# Import generic libcloud functions
-from salt.cloud.libcloudfuncs import *   # pylint: disable=W0614,W0401
+# Import Salt Libs
 import salt.ext.six as six
-try:
-    from salt.utils.openstack import nova
-    HAS_NOVA = True
-except NameError as exc:
-    HAS_NOVA = False
-
-import salt.utils.cloud
-
-# Import nova libs
-try:
-    import novaclient.exceptions
-except ImportError:
-    pass
-
-# Import salt libs
 import salt.utils
 import salt.client
+from salt.utils.openstack import nova
+try:
+    import novaclient.exceptions
+except ImportError as exc:
+    pass
 
-# Import salt.cloud libs
+# Import Salt Cloud Libs
+from salt.cloud.libcloudfuncs import *  # pylint: disable=W0614,W0401
+import salt.utils.cloud
 import salt.utils.pycrypto as sup
 import salt.config as config
 from salt.utils import namespaced_function
@@ -138,7 +226,6 @@ from salt.exceptions import (
     SaltCloudExecutionTimeout
 )
 
-# Import netaddr IP matching
 try:
     from netaddr import all_matching_cidrs
     HAS_NETADDR = True
@@ -149,9 +236,7 @@ except ImportError:
 log = logging.getLogger(__name__)
 request_log = logging.getLogger('requests')
 
-# namespace libcloudfuncs
-get_salt_interface = namespaced_function(get_salt_interface, globals())
-
+__virtualname__ = 'nova'
 
 # Some of the libcloud functions need to be in the same namespace as the
 # functions defined in the module, so we create new function objects inside
@@ -160,13 +245,20 @@ script = namespaced_function(script, globals())
 reboot = namespaced_function(reboot, globals())
 
 
-# Only load in this module is the OPENSTACK configurations are in place
+# Only load in this module if the Nova configurations are in place
 def __virtual__():
     '''
     Check for Nova configurations
     '''
     request_log.setLevel(getattr(logging, __opts__.get('requests_log_level', 'warning').upper()))
-    return HAS_NOVA
+
+    if get_configured_provider() is False:
+        return False
+
+    if get_dependencies() is False:
+        return False
+
+    return __virtualname__
 
 
 def get_configured_provider():
@@ -175,7 +267,22 @@ def get_configured_provider():
     '''
     return config.is_provider_configured(
         __opts__,
-        __active_provider_name__ or 'nova'
+        __active_provider_name__ or __virtualname__,
+        ('user', 'tenant', 'identity_url', 'compute_region',)
+    )
+
+
+def get_dependencies():
+    '''
+    Warn if dependencies aren't met.
+    '''
+    deps = {
+        'netaddr': HAS_NETADDR,
+        'python-novaclient': nova.check_nova(),
+    }
+    return config.check_driver_dependencies(
+        __virtualname__,
+        deps
     )
 
 
@@ -191,6 +298,7 @@ def get_conn():
     kwargs['project_id'] = vm_['tenant']
     kwargs['auth_url'] = vm_['identity_url']
     kwargs['region_name'] = vm_['compute_region']
+    kwargs['use_keystoneauth'] = vm_['use_keystoneauth']
 
     if 'password' in vm_:
         kwargs['password'] = vm_['password']
@@ -225,11 +333,14 @@ def get_image(conn, vm_):
     '''
     Return the image object to use
     '''
-    image_list = conn.image_list()
-
-    vm_image = config.get_cloud_config_value('image', vm_, __opts__).encode(
+    vm_image = config.get_cloud_config_value('image', vm_, __opts__, default='').encode(
         'ascii', 'salt-cloud-force-ascii'
     )
+    if not vm_image:
+        log.debug('No image set, must be boot from volume')
+        return None
+
+    image_list = conn.image_list()
 
     for img in image_list:
         if vm_image in (image_list[img]['id'], img):
@@ -240,13 +351,24 @@ def get_image(conn, vm_):
         return image['id']
     except novaclient.exceptions.NotFound as exc:
         raise SaltCloudNotFound(
-            'The specified image, {0!r}, could not be found: {1}'.format(
+            'The specified image, \'{0}\', could not be found: {1}'.format(
                 vm_image,
                 str(exc)
             )
         )
 
 
+def get_block_mapping_opts(vm_):
+    ret = {}
+    ret['block_device_mapping'] = config.get_cloud_config_value('block_device_mapping', vm_, __opts__, default={})
+    ret['block_device'] = config.get_cloud_config_value('block_device', vm_, __opts__, default=[])
+    ret['ephemeral'] = config.get_cloud_config_value('ephemeral', vm_, __opts__, default=[])
+    ret['swap'] = config.get_cloud_config_value('swap', vm_, __opts__, default=None)
+    ret['snapshot'] = config.get_cloud_config_value('snapshot', vm_, __opts__, default=None)
+    ret['boot_volume'] = config.get_cloud_config_value('boot_volume', vm_, __opts__, default=None)
+    return ret
+
+
 def show_instance(name, call=None):
     '''
     Show the details from the provider concerning an instance
@@ -258,7 +380,7 @@ def show_instance(name, call=None):
 
     conn = get_conn()
     node = conn.show_instance(name).__dict__
-    salt.utils.cloud.cache_node(node, __active_provider_name__, __opts__)
+    __utils__['cloud.cache_node'](node, __active_provider_name__, __opts__)
     return node
 
 
@@ -275,7 +397,7 @@ def get_size(conn, vm_):
         if vm_size and str(vm_size) in (str(sizes[size]['id']), str(size)):
             return sizes[size]['id']
     raise SaltCloudNotFound(
-        'The specified size, {0!r}, could not be found.'.format(vm_size)
+        'The specified size, \'{0}\', could not be found.'.format(vm_size)
     )
 
 
@@ -342,6 +464,28 @@ def rackconnect(vm_):
     )
 
 
+def rackconnectv3(vm_):
+    '''
+    Determine if server is using rackconnectv3 or not
+    Return the rackconnect network name or False
+    '''
+    return config.get_cloud_config_value(
+        'rackconnectv3', vm_, __opts__, default=False,
+        search_global=False
+    )
+
+
+def cloudnetwork(vm_):
+    '''
+    Determine if we should use an extra network to bootstrap
+    Either 'False' (default) or 'True'.
+    '''
+    return config.get_cloud_config_value(
+        'cloudnetwork', vm_, __opts__, default=False,
+        search_global=False
+    )
+
+
 def managedcloud(vm_):
     '''
     Determine if we should wait for the managed cloud automation before
@@ -363,11 +507,12 @@ def destroy(name, conn=None, call=None):
             '-a or --action.'
         )
 
-    salt.utils.cloud.fire_event(
+    __utils__['cloud.fire_event'](
         'event',
         'destroying instance',
         'salt/cloud/{0}/destroying'.format(name),
-        {'name': name},
+        args={'name': name},
+        sock_dir=__opts__['sock_dir'],
         transport=__opts__['transport']
     )
 
@@ -381,10 +526,11 @@ def destroy(name, conn=None, call=None):
     profile = None
     if 'metadata' in node.extra and 'profile' in node.extra['metadata']:
         profile = node.extra['metadata']['profile']
+
     flush_mine_on_destroy = False
-    if profile is not None and profile in profiles:
-        if 'flush_mine_on_destroy' in profiles[profile]:
-            flush_mine_on_destroy = profiles[profile]['flush_mine_on_destroy']
+    if profile and profile in profiles and 'flush_mine_on_destroy' in profiles[profile]:
+        flush_mine_on_destroy = profiles[profile]['flush_mine_on_destroy']
+
     if flush_mine_on_destroy:
         log.info('Clearing Salt Mine: {0}'.format(name))
         salt_client = salt.client.get_local_client(__opts__['conf_file'])
@@ -399,17 +545,19 @@ def destroy(name, conn=None, call=None):
     if ret:
         log.info('Destroyed VM: {0}'.format(name))
         # Fire destroy action
-        salt.utils.cloud.fire_event(
+        __utils__['cloud.fire_event'](
             'event',
             'destroyed instance',
             'salt/cloud/{0}/destroyed'.format(name),
-            {'name': name},
+            args={'name': name},
+            sock_dir=__opts__['sock_dir'],
             transport=__opts__['transport']
         )
         if __opts__.get('delete_sshkeys', False) is True:
             salt.utils.cloud.remove_sshkey(getattr(node, __opts__.get('ssh_interface', 'public_ips'))[0])
         if __opts__.get('update_cachedir', False) is True:
-            salt.utils.cloud.delete_minion_cachedir(name, __active_provider_name__.split(':')[0], __opts__)
+            __utils__['cloud.delete_minion_cachedir'](name, __active_provider_name__.split(':')[0], __opts__)
+        __utils__['cloud.cachedir_index_del'](name)
         return True
 
     log.error('Failed to Destroy VM: {0}'.format(name))
@@ -510,13 +658,20 @@ def request_instance(vm_=None, call=None):
         'config_drive', vm_, __opts__, search_global=False
     )
 
-    salt.utils.cloud.fire_event(
+    kwargs.update(get_block_mapping_opts(vm_))
+
+    __utils__['cloud.fire_event'](
         'event',
         'requesting instance',
         'salt/cloud/{0}/requesting'.format(vm_['name']),
-        {'kwargs': {'name': kwargs['name'],
-                    'image': kwargs['image_id'],
-                    'size': kwargs['flavor_id']}},
+        args={
+            'kwargs': {
+                'name': kwargs['name'],
+                'image': kwargs.get('image_id', 'Boot From Volume'),
+                'size': kwargs['flavor_id'],
+            }
+        },
+        sock_dir=__opts__['sock_dir'],
         transport=__opts__['transport']
     )
 
@@ -530,6 +685,35 @@ def request_instance(vm_=None, call=None):
                 vm_['name'], exc
             )
         )
+    if data.extra.get('password', None) is None and vm_.get('key_filename', None) is None:
+        raise SaltCloudSystemExit('No password returned.  Set ssh_key_file.')
+
+    floating_ip_conf = config.get_cloud_config_value('floating_ip',
+                                                     vm_,
+                                                     __opts__,
+                                                     search_global=False,
+                                                     default={})
+    if floating_ip_conf.get('auto_assign', False):
+        pool = floating_ip_conf.get('pool', 'public')
+        floating_ip = None
+        for fl_ip, opts in six.iteritems(conn.floating_ip_list()):
+            if opts['fixed_ip'] is None and opts['pool'] == pool:
+                floating_ip = fl_ip
+                break
+        if floating_ip is None:
+            floating_ip = conn.floating_ip_create(pool)['ip']
+        try:
+            conn.floating_ip_associate(kwargs['name'], floating_ip)
+            vm_['floating_ip'] = floating_ip
+        except Exception as exc:
+            raise SaltCloudSystemExit(
+                'Error assigning floating_ip for {0} on Nova\n\n'
+                'The following exception was thrown by libcloud when trying to '
+                'assing a floating ip: {1}\n'.format(
+                    vm_['name'], exc
+                )
+            )
+
     vm_['password'] = data.extra.get('password', '')
 
     return data, vm_
@@ -539,28 +723,44 @@ def create(vm_):
     '''
     Create a single VM from a data dict
     '''
+    try:
+        # Check for required profile parameters before sending any API calls.
+        if vm_['profile'] and config.is_profile_configured(__opts__,
+                                                           __active_provider_name__ or 'nova',
+                                                           vm_['profile'],
+                                                           vm_=vm_) is False:
+            return False
+    except AttributeError:
+        pass
+
     deploy = config.get_cloud_config_value('deploy', vm_, __opts__)
     key_filename = config.get_cloud_config_value(
         'ssh_key_file', vm_, __opts__, search_global=False, default=None
     )
     if key_filename is not None and not os.path.isfile(key_filename):
         raise SaltCloudConfigError(
-            'The defined ssh_key_file {0!r} does not exist'.format(
+            'The defined ssh_key_file \'{0}\' does not exist'.format(
                 key_filename
             )
         )
 
     vm_['key_filename'] = key_filename
 
-    salt.utils.cloud.fire_event(
+    # Since using "provider: <provider-engine>" is deprecated, alias provider
+    # to use driver: "driver: <provider-engine>"
+    if 'provider' in vm_:
+        vm_['driver'] = vm_.pop('provider')
+
+    __utils__['cloud.fire_event'](
         'event',
         'starting create',
         'salt/cloud/{0}/creating'.format(vm_['name']),
-        {
+        args={
             'name': vm_['name'],
             'profile': vm_['profile'],
-            'provider': vm_['provider'],
+            'provider': vm_['driver'],
         },
+        sock_dir=__opts__['sock_dir'],
         transport=__opts__['transport']
     )
     conn = get_conn()
@@ -569,7 +769,7 @@ def create(vm_):
         # This was probably created via another process, and doesn't have
         # things like salt keys created yet, so let's create them now.
         if 'pub_key' not in vm_ and 'priv_key' not in vm_:
-            log.debug('Generating minion keys for {0[name]!r}'.format(vm_))
+            log.debug('Generating minion keys for \'{0[name]}\''.format(vm_))
             vm_['priv_key'], vm_['pub_key'] = salt.utils.cloud.gen_keys(
                 salt.config.get_cloud_config_value(
                     'keysize',
@@ -614,26 +814,11 @@ def create(vm_):
             # Still not running, trigger another iteration
             return
 
-        rackconnectv3 = config.get_cloud_config_value(
-            'rackconnectv3', vm_, __opts__, default=False,
-            search_global=False
-        )
-
-        if rackconnectv3:
-            networkname = rackconnectv3
-            for network in node['addresses'].get(networkname, []):
-                if network['version'] is 4:
-                    node['extra']['access_ip'] = network['addr']
-                    break
-            vm_['rackconnect'] = True
-
         if rackconnect(vm_) is True:
             extra = node.get('extra', {})
             rc_status = extra.get('metadata', {}).get(
                 'rackconnect_automation_status', '')
-            access_ip = extra.get('access_ip', '')
-
-            if rc_status != 'DEPLOYED' and not rackconnectv3:
+            if rc_status != 'DEPLOYED':
                 log.debug('Waiting for Rackconnect automation to complete')
                 return
 
@@ -648,25 +833,63 @@ def create(vm_):
                 log.debug('Waiting for managed cloud automation to complete')
                 return
 
+        access_ip = node.get('extra', {}).get('access_ip', '')
+
+        rcv3 = rackconnectv3(vm_) in node['addresses']
+        sshif = ssh_interface(vm_) in node['addresses']
+
+        if any((rcv3, sshif)):
+            networkname = rackconnectv3(vm_) if rcv3 else ssh_interface(vm_)
+            for network in node['addresses'].get(networkname, []):
+                if network['version'] is 4:
+                    access_ip = network['addr']
+                    break
+            vm_['cloudnetwork'] = True
+
+        # Conditions to pass this
+        #
+        #     Rackconnect v2: vm_['rackconnect'] = True
+        #         If this is True, then the server will not be accessible from the ipv4 addres in public_ips.
+        #         That interface gets turned off, and an ipv4 from the dedicated firewall is routed to the
+        #         server.  In this case we can use the private_ips for ssh_interface, or the access_ip.
+        #
+        #     Rackconnect v3: vm['rackconnectv3'] = <cloudnetwork>
+        #         If this is the case, salt will need to use the cloud network to login to the server.  There
+        #         is no ipv4 address automatically provisioned for these servers when they are booted.  SaltCloud
+        #         also cannot use the private_ips, because that traffic is dropped at the hypervisor.
+        #
+        #     CloudNetwork: vm['cloudnetwork'] = True
+        #         If this is True, then we should have an access_ip at this point set to the ip on the cloud
+        #         network.  If that network does not exist in the 'addresses' dictionary, then SaltCloud will
+        #         use the initial access_ip, and not overwrite anything.
+
+        if any((cloudnetwork(vm_), rackconnect(vm_))) and (ssh_interface(vm_) != 'private_ips' or rcv3) and access_ip != '':
+            data.public_ips = [access_ip, ]
+            return data
+
         result = []
 
         if 'private_ips' not in node and 'public_ips' not in node and \
+           'floating_ips' not in node and 'fixed_ips' not in node and \
            'access_ip' in node.get('extra', {}):
             result = [node['extra']['access_ip']]
 
         private = node.get('private_ips', [])
         public = node.get('public_ips', [])
+        fixed = node.get('fixed_ips', [])
+        floating = node.get('floating_ips', [])
+
         if private and not public:
-            log.warn(
+            log.warning(
                 'Private IPs returned, but not public... Checking for '
                 'misidentified IPs'
             )
             for private_ip in private:
                 private_ip = preferred_ip(vm_, [private_ip])
                 if salt.utils.cloud.is_public_ip(private_ip):
-                    log.warn('{0} is a public IP'.format(private_ip))
+                    log.warning('{0} is a public IP'.format(private_ip))
                     data.public_ips.append(private_ip)
-                    log.warn(
+                    log.warning(
                         (
                             'Public IP address was not ready when we last'
                             ' checked.  Appending public IP address now.'
@@ -674,14 +897,40 @@ def create(vm_):
                     )
                     public = data.public_ips
                 else:
-                    log.warn('{0} is a private IP'.format(private_ip))
+                    log.warning('{0} is a private IP'.format(private_ip))
                     ignore_ip = ignore_cidr(vm_, private_ip)
                     if private_ip not in data.private_ips and not ignore_ip:
                         result.append(private_ip)
 
-        if rackconnect(vm_) is True:
-            if ssh_interface(vm_) != 'private_ips' or rackconnectv3:
-                data.public_ips = access_ip
+        # populate return data with private_ips
+        # when ssh_interface is set to private_ips and public_ips exist
+        if not result and ssh_interface(vm_) == 'private_ips':
+            for private_ip in private:
+                ignore_ip = ignore_cidr(vm_, private_ip)
+                if private_ip not in data.private_ips and not ignore_ip:
+                    result.append(private_ip)
+
+        non_private_ips = []
+
+        if public:
+            data.public_ips = public
+            if ssh_interface(vm_) == 'public_ips':
+                non_private_ips.append(public)
+
+        if floating:
+            data.floating_ips = floating
+            if ssh_interface(vm_) == 'floating_ips':
+                non_private_ips.append(floating)
+
+        if fixed:
+            data.fixed_ips = fixed
+            if ssh_interface(vm_) == 'fixed_ips':
+                non_private_ips.append(fixed)
+
+        if non_private_ips:
+            log.debug('result = {0}'.format(non_private_ips))
+            data.private_ips = result
+            if ssh_interface(vm_) != 'private_ips':
                 return data
 
         if result:
@@ -690,11 +939,6 @@ def create(vm_):
             if ssh_interface(vm_) == 'private_ips':
                 return data
 
-        if public:
-            data.public_ips = public
-            if ssh_interface(vm_) != 'private_ips':
-                return data
-
     try:
         data = salt.utils.cloud.wait_for_ip(
             __query_node_data,
@@ -717,17 +961,23 @@ def create(vm_):
 
     if ssh_interface(vm_) == 'private_ips':
         ip_address = preferred_ip(vm_, data.private_ips)
-    elif rackconnect(vm_) is True and ssh_interface(vm_) != 'private_ips':
-        ip_address = data.public_ips
+    elif ssh_interface(vm_) == 'fixed_ips':
+        ip_address = preferred_ip(vm_, data.fixed_ips)
+    elif ssh_interface(vm_) == 'floating_ips':
+        ip_address = preferred_ip(vm_, data.floating_ips)
     else:
         ip_address = preferred_ip(vm_, data.public_ips)
     log.debug('Using IP address {0}'.format(ip_address))
 
-    if get_salt_interface(vm_) == 'private_ips':
+    if salt.utils.cloud.get_salt_interface(vm_, __opts__) == 'private_ips':
         salt_ip_address = preferred_ip(vm_, data.private_ips)
         log.info('Salt interface set to: {0}'.format(salt_ip_address))
-    elif rackconnect(vm_) is True and get_salt_interface(vm_) != 'private_ips':
-        salt_ip_address = data.public_ips
+    elif salt.utils.cloud.get_salt_interface(vm_, __opts__) == 'fixed_ips':
+        salt_ip_address = preferred_ip(vm_, data.fixed_ips)
+        log.info('Salt interface set to: {0}'.format(salt_ip_address))
+    elif salt.utils.cloud.get_salt_interface(vm_, __opts__) == 'floating_ips':
+        salt_ip_address = preferred_ip(vm_, data.floating_ips)
+        log.info('Salt interface set to: {0}'.format(salt_ip_address))
     else:
         salt_ip_address = preferred_ip(vm_, data.public_ips)
         log.debug('Salt interface set to: {0}'.format(salt_ip_address))
@@ -738,32 +988,40 @@ def create(vm_):
     vm_['ssh_host'] = ip_address
     vm_['salt_host'] = salt_ip_address
 
-    ret = salt.utils.cloud.bootstrap(vm_, __opts__)
+    ret = __utils__['cloud.bootstrap'](vm_, __opts__)
 
     ret.update(data.__dict__)
 
     if 'password' in ret['extra']:
         del ret['extra']['password']
 
-    log.info('Created Cloud VM {0[name]!r}'.format(vm_))
+    log.info('Created Cloud VM \'{0[name]}\''.format(vm_))
     log.debug(
-        '{0[name]!r} VM creation details:\n{1}'.format(
-            vm_, pprint.pformat(data)
+        '\'{0[name]}\' VM creation details:\n{1}'.format(
+            vm_, pprint.pformat(data.__dict__)
         )
     )
 
-    salt.utils.cloud.fire_event(
+    event_data = {
+        'name': vm_['name'],
+        'profile': vm_['profile'],
+        'provider': vm_['driver'],
+        'instance_id': vm_['instance_id'],
+        'floating_ips': data.floating_ips,
+        'fixed_ips': data.fixed_ips,
+        'private_ips': data.private_ips,
+        'public_ips': data.public_ips
+    }
+
+    __utils__['cloud.fire_event'](
         'event',
         'created instance',
         'salt/cloud/{0}/created'.format(vm_['name']),
-        {
-            'name': vm_['name'],
-            'profile': vm_['profile'],
-            'provider': vm_['provider'],
-        },
+        args=event_data,
+        sock_dir=__opts__['sock_dir'],
         transport=__opts__['transport']
     )
-
+    __utils__['cloud.cachedir_index_add'](vm_['name'], vm_['profile'], 'nova', vm_['driver'])
     return ret
 
 
@@ -799,7 +1057,11 @@ def list_nodes(call=None, **kwargs):
     if not server_list:
         return {}
     for server in server_list:
-        server_tmp = conn.server_show(server_list[server]['id'])[server]
+        server_tmp = conn.server_show(server_list[server]['id']).get(server)
+
+        # If the server is deleted while looking it up, skip
+        if server_tmp is None:
+            continue
 
         private = []
         public = []
@@ -859,10 +1121,30 @@ def list_nodes_full(call=None, **kwargs):
         except IndexError as exc:
             ret = {}
 
-    salt.utils.cloud.cache_node_list(ret, __active_provider_name__.split(':')[0], __opts__)
+    __utils__['cloud.cache_node_list'](ret, __active_provider_name__.split(':')[0], __opts__)
     return ret
 
 
+def list_nodes_min(call=None, **kwargs):
+    '''
+    Return a list of the VMs that in this location
+    '''
+    if call == 'action':
+        raise SaltCloudSystemExit(
+            (
+                'The list_nodes_min function must be called with'
+                ' -f or --function.'
+            )
+        )
+
+    conn = get_conn()
+    server_list = conn.server_list_min()
+
+    if not server_list:
+        return {}
+    return server_list
+
+
 def list_nodes_select(call=None):
     '''
     Return a list of the VMs that are on the provider, with select fields
@@ -1026,3 +1308,111 @@ def virtual_interface_create(name, net_name, **kwargs):
     '''
     conn = get_conn()
     return conn.virtual_interface_create(name, net_name)
+
+
+def floating_ip_pool_list(call=None):
+    '''
+    List all floating IP pools
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'function':
+        raise SaltCloudSystemExit(
+            'The floating_ip_pool_list action must be called with -f or --function'
+        )
+
+    conn = get_conn()
+    return conn.floating_ip_pool_list()
+
+
+def floating_ip_list(call=None):
+    '''
+    List floating IPs
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'function':
+        raise SaltCloudSystemExit(
+            'The floating_ip_list action must be called with -f or --function'
+        )
+
+    conn = get_conn()
+    return conn.floating_ip_list()
+
+
+def floating_ip_create(kwargs, call=None):
+    '''
+    Allocate a floating IP
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'function':
+        raise SaltCloudSystemExit(
+            'The floating_ip_create action must be called with -f or --function'
+        )
+
+    if 'pool' not in kwargs:
+        log.error('pool is required')
+        return False
+
+    conn = get_conn()
+    return conn.floating_ip_create(kwargs['pool'])
+
+
+def floating_ip_delete(kwargs, call=None):
+    '''
+    De-allocate floating IP
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'function':
+        raise SaltCloudSystemExit(
+            'The floating_ip_delete action must be called with -f or --function'
+        )
+
+    if 'floating_ip' not in kwargs:
+        log.error('floating_ip is required')
+        return False
+
+    conn = get_conn()
+    return conn.floating_ip_delete(kwargs['floating_ip'])
+
+
+def floating_ip_associate(name, kwargs, call=None):
+    '''
+    Associate a floating IP address to a server
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'action':
+        raise SaltCloudSystemExit(
+            'The floating_ip_associate action must be called with -a of --action.'
+        )
+
+    if 'floating_ip' not in kwargs:
+        log.error('floating_ip is required')
+        return False
+
+    conn = get_conn()
+    conn.floating_ip_associate(name, kwargs['floating_ip'])
+    return list_nodes()[name]
+
+
+def floating_ip_disassociate(name, kwargs, call=None):
+    '''
+    Disassociate a floating IP from a server
+
+    .. versionadded:: 2016.3.0
+    '''
+    if call != 'action':
+        raise SaltCloudSystemExit(
+            'The floating_ip_disassociate action must be called with -a of --action.'
+        )
+
+    if 'floating_ip' not in kwargs:
+        log.error('floating_ip is required')
+        return False
+
+    conn = get_conn()
+    conn.floating_ip_disassociate(name, kwargs['floating_ip'])
+    return list_nodes()[name]
